module Distributed.Model.Ring.ChangRoberts where

import Distributed.Model.Ring
import Distributed.Model.Semantics

import Data.Random.Source.Std

data CRState id =
    Follower id -- ^ The highest ID a node has seen if it's not the leader
  | Leader -- ^ The node believes it is currently the leader
  deriving (Eq, Ord, Show)

data CRMessage id =
    MyLeader id
  deriving (Eq, Ord, Show)

changRoberts :: Ord id => id -> RingT (CRMessage id) IO (CRState id)
changRoberts nId = do
  sendLeft $ MyLeader nId
  go Leader
  where
    go (Follower ourLeader) = do
      (Msg peer (MyLeader otherLeader)) <- recvRing
      case peer of
        L -> dropMessage
        R -> do
          let
            newLeader = max ourLeader otherLeader
          sendLeft $ MyLeader newLeader
          go $ Follower newLeader
    go Leader = do
      (Msg peer (MyLeader otherLeader)) <- recvRing
      case peer of
        L -> dropMessage
        R ->
          case nId `compare` otherLeader of
            LT -> do -- This node isn't the leader any more
              sendLeft $ MyLeader otherLeader
              go $ Follower otherLeader
            EQ -> -- This node is the leader
              return $ Leader
            GT -> -- This node is still the leader
              go $ Leader
    dropMessage =
      error "Received a message from left peer somehow"

runChangRoberts :: Ord id => [id] -> IO [Maybe (CRState id)]
runChangRoberts nodes =
  runRing (changRoberts <$> nodes) StdRandom